;; Generated package description from plz-event-source.el  -*- no-byte-compile: t -*-
(define-package "plz-event-source" "0.1.2pre0.20240924.181709" "Plz Event Source" '((emacs "26.3") (plz-media-type "0.1.0")) :commit "1d584eb8d015faa839141ac68e934dde89a6e2ab" :authors '(("r0man" . "roman@burningswell.com")) :maintainer '("r0man" . "roman@burningswell.com") :keywords '("comm" "network" "http") :url "https://github.com/r0man/plz-event-source")
